/**
 * @hidden
 */
declare function load_native_lib(): any;
/**
 * @hidden
 */
declare let native: any;
/**
 * @hidden
 */
declare class Component {
    protected comp: any;
    constructor(comp: any);
    run(): void;
    show(): void;
    hide(): void;
    get window(): Window;
    send_mouse_click(x: number, y: number): void;
    send_keyboard_string_sequence(s: String): void;
}
interface Window {
    show(): void;
    hide(): void;
}
/**
 * @hidden
 */
interface Callback {
    (): any;
    setHandler(cb: any): void;
}
/**
 * ModelPeer is the interface that the run-time implements. An instance is
 * set on dynamic Model<T> instances and can be used to notify the run-time
 * of changes in the structure or data of the model.
 */
interface ModelPeer {
    /**
     * Call this function from our own model to notify that fields of data
     * in the specified row have changed.
     * @argument row
     */
    rowDataChanged(row: number): void;
    /**
     * Call this function from your own model to notify that one or multiple
     * rows were added to the model, starting at the specified row.
     * @param row
     * @param count
     */
    rowAdded(row: number, count: number): void;
    /**
     * Call this function from your own model to notify that one or multiple
     * rows were removed from the model, starting at the specified row.
     * @param row
     * @param count
     */
    rowRemoved(row: number, count: number): void;
}
/**
 * Model<T> is the interface for feeding dynamic data into
 * `.60` views.
 *
 * A model is organized like a table with rows of data. The
 * fields of the data type T behave like columns.
 */
interface Model<T> {
    /**
     * Implementations of this function must return the current number of rows.
     */
    rowCount(): number;
    /**
     * Implementations of this function must return the data at the specified row.
     * @param row
     */
    rowData(row: number): T;
    /**
     * Implementations of this function must store the provided data parameter
     * in the model at the specified row.
     * @param row
     * @param data
     */
    setRowData(row: number, data: T): void;
    /**
     * This public member is set by the run-time and implementation must use this
     * to notify the run-time of changes in the model.
     */
    notify: ModelPeer;
}
/**
 * @hidden
 */
declare class NullPeer implements ModelPeer {
    rowDataChanged(row: number): void;
    rowAdded(row: number, count: number): void;
    rowRemoved(row: number, count: number): void;
}
/**
 * ArrayModel wraps a JavaScript array for use in `.60` views. The underlying
 * array can be modified with the [[ArrayModel.push]] and [[ArrayModel.remove]] methods.
 */
declare class ArrayModel<T> implements Model<T> {
    /**
     * @hidden
     */
    private a;
    notify: ModelPeer;
    /**
     * Creates a new ArrayModel.
     *
     * @param arr
     */
    constructor(arr: Array<T>);
    rowCount(): number;
    rowData(row: number): T;
    setRowData(row: number, data: T): void;
    /**
     * Pushes new values to the array that's backing the model and notifies
     * the run-time about the added rows.
     * @param values
     */
    push(...values: T[]): void;
    /**
     * Removes the specified number of element from the array that's backing
     * the model, starting at the specified index. This is equivalent to calling
     * Array.slice() on the array and notifying the run-time about the removed
     * rows.
     * @param index
     * @param size
     */
    remove(index: number, size: number): void;
    get length(): number;
    values(): IterableIterator<T>;
    entries(): IterableIterator<[number, T]>;
}
